#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>
#include <string.h>
#include <dotconf.h>
#include <ctype.h>
#include <sched.h>

#include "macros.h"
#include "common.h"
#include "xmalloc.h"
#include "list.h"
#include "str.h"
#include "buffer.h"
#include "misc.h"

#include "hal-mgr-conf.h"

#define TIMEOUT_MIN		0
#define	TIMEOUT_MAX		60000

#define CONTEXT_ROOT		0x1
#define CONTEXT_PISERVER	0x2
#define CONTEXT_PARAMETERS	0x4

#define HAVE_OPTION_RETRYTIMER	0x01
#define HAVE_OPTION_PRIORITY	0x02
#define	HAVE_OPTION_MGR_SOCKET	0x04

#define warn(fmt, arg...) fprintf(stderr, "warn: %s: "fmt CRLF, \
					__FUNCTION__, ##arg)

/* List of all PI servers, parameters and options. */
extern struct slist *pi_srv_options_list;

struct context_info {
	char			*end_token;	/* pointer to section end token */
	unsigned int		context;	/* context flags */
	char			mgr_socketfile[PATH_MAX];
	char			srv_socketfile[PATH_MAX];
	char			inet_addr[PATH_MAX];
	char			name[NAME_MAX];
	int			side;		/* master or slave ? */
	int			priority;
	int			retry_tout_msec;
	int			timeout;
	str_t			errstr;
	int			option_flags;
};

static char end_piserver[] = "</piserver>";

static const char *opt_priority(command_t *cmd, context_t *ctx);
static const char *opt_retrytimer(command_t *cmd, context_t *ctx);
static const char *opt_mgr_socket(command_t *cmd, context_t *ctx);
static const char *open_piserver(command_t *cmd, context_t *ctx);
static const char *close_piserver(command_t *cmd, context_t *ctx);
static const char *opt_name(command_t *cmd, context_t *ctx);
static const char *opt_side(command_t *cmd, context_t *ctx);
static const char *opt_srv_socket(command_t *cmd, context_t *ctx);
static const char *opt_inet_addr(command_t *cmd, context_t *ctx);
static const char *opt_timeout(command_t *cmd, context_t *ctx);

static configoption_t options[] = {
	{ "priority",	ARG_INT,  opt_priority,	  NULL, CONTEXT_ROOT },
	{ "retrytimer", ARG_INT,  opt_retrytimer, NULL, CONTEXT_ROOT },
	{ "socket_bind",ARG_STR,  opt_mgr_socket, NULL, CONTEXT_ROOT },
	{ "<piserver>", ARG_NONE, open_piserver,  NULL, CONTEXT_ROOT },
	{ end_piserver, ARG_NONE, close_piserver, NULL, CONTEXT_PISERVER },
	{ "name",	ARG_STR,  opt_name, 	  NULL, CONTEXT_PISERVER },
	{ "side",	ARG_STR,  opt_side, 	  NULL, CONTEXT_PISERVER },
	{ "socket",	ARG_STR,  opt_srv_socket, NULL, CONTEXT_PISERVER },
	{ "inet_addr",	ARG_STR,  opt_inet_addr,  NULL, CONTEXT_PISERVER },
	{ "timeout",	ARG_INT,  opt_timeout,    NULL, CONTEXT_PISERVER },
	LAST_OPTION
};

static const char *opt_priority(command_t *cmd, context_t *opaque)
{
	struct context_info *ctx = (struct context_info *)opaque;

	if (ctx->option_flags & HAVE_OPTION_PRIORITY)
		return "Duplicated option `priority'";

	if (cmd->data.value < sched_get_priority_min(SCHED_RR) ||
	    cmd->data.value > sched_get_priority_max(SCHED_RR)) {
		str_reset(ctx->errstr);
		str_snprintf(ctx->errstr, "Option `priority' range is %d..%d.",
					sched_get_priority_min(SCHED_RR),
					sched_get_priority_max(SCHED_RR));
		return str_ptr(ctx->errstr);
	}

	ctx->priority = cmd->data.value;
	ctx->option_flags |= HAVE_OPTION_PRIORITY;

	return NULL;
}

static const char *opt_retrytimer(command_t *cmd, context_t *opaque)
{
	struct context_info *ctx = (struct context_info *)opaque;

	if (ctx->option_flags & HAVE_OPTION_RETRYTIMER)
		return "Duplicated option `retrytimer'";

	ctx->retry_tout_msec = cmd->data.value * 1000;
	ctx->option_flags |= HAVE_OPTION_RETRYTIMER;

	return NULL;
}

static const char *opt_mgr_socket(command_t *cmd, context_t *opaque)
{
	struct context_info *ctx = (struct context_info *)opaque;

	if (ctx->option_flags & HAVE_OPTION_MGR_SOCKET)
		return "Duplicated option `socket'";

	xstrncpy(ctx->mgr_socketfile, cmd->data.str, sizeof(ctx->mgr_socketfile));
	ctx->option_flags |= HAVE_OPTION_MGR_SOCKET;

	return NULL;
}

/* open_piserver - handler opens <piserver> section */
static const char *open_piserver(command_t *cmd, context_t *opaque)
{
	struct context_info *ctx;
	struct pi_srv_options *opts;
	struct slist *list;
	char *last_end_token;
	const char *err = NULL;

	ctx = (struct context_info *)opaque;

	if (ctx->context & CONTEXT_PISERVER)
		return "Nesting section <piserver> not allowed.";

	ctx->context |= CONTEXT_PISERVER;
	last_end_token = ctx->end_token;
	ctx->end_token = end_piserver;

	while (!cmd->configfile->eof) {
		err = dotconf_command_loop_until_error(cmd->configfile);
		if (err) {
			if (err == end_piserver)
				break;
			dotconf_warning(cmd->configfile, DCLOG_ERR, 0, "%s", err);
		} else {
			return "Section <piserver> is not terminated properly </piserver > is missed";
		}
	}
	/* Check mandatory options presense. */
	if (strlen(ctx->name) == 0)
		dotconf_warning(cmd->configfile, DCLOG_WARNING, 0,
			"Missing `name' in section <piserver>.");
	if ((strlen(ctx->srv_socketfile) == 0) && (strlen(ctx->inet_addr) == 0))
		dotconf_warning(cmd->configfile, DCLOG_WARNING, 0,
			"Missing `socket' or `inet_addr' in section <piserver>.");
	if ((int)ctx->side < 0)
		dotconf_warning(cmd->configfile, DCLOG_WARNING, 0,
			"Must specify `side' option in section <piserver>.");

	/* Add option element to the list if OK. */
	if (strlen(ctx->name) && (strlen(ctx->srv_socketfile) || strlen(ctx->inet_addr)) &&
	    ctx->side >= 0) {
		/* All mandatory options are present. Create it. */
		opts = xmalloc0(sizeof(*opts));
		opts->timeout = ctx->timeout > 0 ? ctx->timeout : 0;
		xstrncpy(opts->name, ctx->name, sizeof(opts->name));
		xstrncpy(opts->socketfile, ctx->srv_socketfile, sizeof(opts->socketfile));
		xstrncpy(opts->inet_addr, ctx->inet_addr, sizeof(opts->inet_addr));
		opts->side = ctx->side;
		/* Add options to the list. */
		list = slist_new();
		list->data = opts;
		pi_srv_options_list = slist_add(pi_srv_options_list, list);
	}

	/* Reset options to defaults. */
	ctx->timeout = -1;
	ctx->side = -1;
	ctx->name[0] = ctx->srv_socketfile[0] = ctx->inet_addr[0] = 0;
	/* Section was terminated? */
	if (err != end_piserver)
		return err;
	/* Quit section. */
	ctx->context &= ~CONTEXT_PISERVER;
	ctx->end_token = last_end_token;
	return NULL;
}

/* close_piserver - close <piserver> section by </piserver> */
static const char *close_piserver(command_t *cmd, context_t *opaque)
{
	struct context_info *ctx = (struct context_info *) opaque;

	if ((ctx->context & CONTEXT_PISERVER) == 0)
		return "Token </piserver> is not closing anything.";

	return end_piserver;
}

/*
 * This helper checks that the string is a valid symbol
 * name and has the first char belonging to the set [a-zA-Z]
 * and the following chars belonging to the set [0-9a-zA-Z-_].
 */
static int is_symbol(const char *str)
{
	if (str == NULL)
		return 0;

	if (*str && isalpha(*str)) {
		str++;
		while (*str && (isalnum(*str) ||
		    *str == '-' || *str == '_'))
			str++;
	}
	return *str == 0 ? 1 : 0;
}

static const char *opt_name(command_t *cmd, context_t *opaque)
{
	struct context_info *ctx = (struct context_info *) opaque;

	if (strlen(ctx->name) == 0) {
		if (is_symbol(cmd->data.str)) {
			xstrncpy(ctx->name, cmd->data.str, sizeof(ctx->name));
		} else {
			str_reset(ctx->errstr);
			str_snprintf(ctx->errstr, "Option `name' doesn't look as a valid name.");
			return str_ptr(ctx->errstr);
		}
	} else {
		return "Duplicated option `name' in section <piserver>.";
	}

	return NULL;
}

static const char *opt_side(command_t *cmd, context_t *opaque)
{
	struct context_info *ctx = (struct context_info *) opaque;

	if (strcmp(cmd->data.str, "slave") == 0) {
	//	ctx->side = PI_SIDE_SLAVE;
	} else if (strcmp(cmd->data.str, "master") == 0) {
	//	ctx->side = PI_SIDE_MASTER;
	} else {
		str_reset(ctx->errstr);
		str_snprintf(ctx->errstr, "Option `side' has unrecognized argument.");
		return str_ptr(ctx->errstr);
	}

	ctx->side = 10;

	return NULL;
}

static const char *opt_timeout(command_t *cmd, context_t *opaque)
{
	struct context_info *ctx = (struct context_info *)opaque;

	if (ctx->timeout < 0)
		ctx->timeout = cmd->data.value;
	else
		return "Duplicated option `timeout' in section <piserver>.";

	return NULL;
}


static const char *opt_srv_socket(command_t *cmd, context_t *opaque)
{
	struct context_info *ctx = (struct context_info *) opaque;

	if (strlen(ctx->inet_addr) != 0)
		return "`socket' and `inet_addr' options are mutualy exclusive.";

	if (strlen(ctx->srv_socketfile) == 0)
		xstrncpy(ctx->srv_socketfile, cmd->data.str, sizeof(ctx->srv_socketfile));
	else
		return "Duplicated option `socket' in section <piserver>.";

	return NULL;
}

static const char *opt_inet_addr(command_t *cmd, context_t *opaque)
{
	struct context_info *ctx = (struct context_info *) opaque;

	if (strlen(ctx->srv_socketfile) != 0)
		return "`socket' and `inet_addr' options are mutualy exclusive.";

	if (strlen(ctx->inet_addr) == 0)
		xstrncpy(ctx->inet_addr, cmd->data.str, sizeof(ctx->inet_addr));
	else
		return "Duplicated option `inet_addr' in section <piserver>.";

	return NULL;
}

static const int error_handler(configfile_t *config, int type, long dc_errno, const char *msg)
{
	warn("Parser: %s: file '%s' line %ld", msg, config->filename, config->line);
	return 0;
}

static const char *context_checker(command_t *cmd, unsigned long mask)
{
	struct context_info *ctx = (struct context_info *) cmd->context;

	/* option is OK in root context? */
	if ((mask & CONTEXT_ROOT) && ctx->context == 0)
		return NULL;

	/* otherwise context with option's mask */
	if ((ctx->context & mask) == 0) {
		str_reset(ctx->errstr);
		str_snprintf(ctx->errstr, "Option `%s' is not allowed in this context.", cmd->name);
		return str_ptr(ctx->errstr);
	}

	return NULL;
}

void config_free()
{
	struct slist *list;

	for (list = pi_srv_options_list; list != NULL; list = list->next) {
		struct pi_srv_options *opts = (struct pi_srv_options *)list->data;
		xfree(opts);
	}

	if (pi_srv_options_list != NULL)
		slist_free_all(&pi_srv_options_list, NULL);
}

int config_read(const char *configfile, struct hal_mgr_options *opts)
{
	const char *err = NULL;
	struct context_info *ctx;
	configfile_t *config;

	if (!opts)
		return -1;

	ctx = xmalloc0(sizeof(*ctx));
	ctx->context = 0;
	ctx->end_token = NULL;
	/* reset options to initial state */
	ctx->srv_socketfile[0] = 0;
	ctx->inet_addr[0] = 0;
	ctx->name[0] = 0;
	ctx->timeout = -1;
	ctx->side = -1;
	ctx->errstr = str_new();
	/* reset options to undefined state */
	ctx->priority = -1;
	ctx->retry_tout_msec = -1;
	ctx->option_flags = 0;

	config = dotconf_create((char *)configfile, options, (context_t *)ctx,
					DONT_SUBSTITUTE|DUPLICATE_OPTION_NAMES);
	if (config == NULL) {
		warn("Configuration parser can't create dotconf context.");
		str_free(&ctx->errstr);
		xfree(ctx);
		return -1;
	}

	config->contextchecker = context_checker;
	config->errorhandler = (dotconf_errorhandler_t)error_handler;

	while (!config->eof) {
		err = dotconf_command_loop_until_error(config);
		if (err != NULL)
			dotconf_warning(config, DCLOG_WARNING, 0, "%s", err);
		else
			break;
	}

	/*
	 * Reorder all elements in the list. They must appear
	 * in the same order as in the configuration file.
	 */
	pi_srv_options_list = slist_reverse(pi_srv_options_list);

	opts->retry_tout_msec = ctx->retry_tout_msec;
	opts->priority = ctx->priority;
	xstrncpy(opts->socketfile, ctx->mgr_socketfile, sizeof(opts->socketfile));

	dotconf_cleanup(config);
	str_free(&ctx->errstr);
	xfree(ctx);

	return 0;
}

